Attribute VB_Name = "pAPUMidi"
'Partially emulates the Audio Processing Unit (APU). Sometimes it works.

Option Explicit
DefLng A-Z

Private Tones(3)
Private Volume(3)
Private LastFrame(3)

Public doSound As Boolean

Private vlengths(31) As Long

Public toneAdjust As Double
Public ChannelWrite(3) As Boolean

Private Sub fillArray(a() As Long, ParamArray b() As Variant)
    Dim i
    For i = 0 To UBound(a)
        a(i) = b(i)
    Next i
End Sub

Public Sub pAPUinit()
    'Lookup table used by nester.
    fillArray vlengths, 5, 127, 10, 1, 19, 2, 40, 3, 80, 4, 30, 5, 7, 6, 13, 7, 6, 8, 12, 9, 24, 10, 48, 11, 96, 12, 36, 13, 8, 14, 16, 15
    
    SelectInstrument 0, 80 'Square wave
    SelectInstrument 1, 80 'Square wave
    SelectInstrument 2, 74 'Triangle wave. Used recorder (like a flute
    SelectInstrument 3, 127 'Noise. Used gunshot. Poor but sometimes works.
End Sub

Public Sub playTone(channel, tone, v)
    If tone <> Tones(channel) Or v < Volume(channel) - 3 Or v > Volume(channel) Or v = 0 Then
        stopTone channel
        If doSound And tone <> 0 And v > 0 Then
            Volume(channel) = v
            Tones(channel) = tone
            ToneOn channel, tone, v * 127 / 15
        End If
    End If
End Sub

Public Sub stopTone(channel)
    If Tones(channel) <> 0 Then
        ToneOff channel, Tones(channel)
        Tones(channel) = 0
        Volume(channel) = 0
    End If
End Sub

'Calculates a midi tone given an nes frequency.
'Frequency passed is actual interval in 1/65536's of a second (I hope)
Public Function getTone(ByVal freq) As Long
    If freq <= 0 Then Exit Function
    
    Dim t
    
    freq = 65536 / freq
    t = Log(freq / 8.176) / Log(1.059463)
    
    If t < 1 Then t = 1
    If t > 127 Then t = 127
    getTone = t
End Function

Public Sub PlayRect(ch)
    Dim f, l, v
    If SoundCtrl And pow2(ch) Then
        v = (Sound(ch * 4 + 0) And 15) 'Get volume
        l = vlengths(Sound(ch * 4 + 3) \ 8) 'Get length
        If v > 0 Then
            f = Sound(ch * 4 + 2) + (Sound(ch * 4 + 3) And 7) * 256 'Get frequency
            If f > 1 Then
                If ChannelWrite(ch) Then 'Ensures that a note doesn't replay unless memory written
                    ChannelWrite(ch) = False
                    LastFrame(ch) = Frames + l
                    playTone ch, getTone(f), v
                End If
            Else
                stopTone ch
            End If
        Else
            stopTone ch
        End If
    Else
        ChannelWrite(ch) = True
        stopTone ch
    End If
    If Frames >= LastFrame(ch) Then
        stopTone ch
    End If
End Sub

Public Sub PlayTriangle(ch)
    Dim f, l, v
    If SoundCtrl And pow2(ch) Then
        v = 9 'triangle
        l = vlengths(Sound(ch * 4 + 3) \ 8)
        If v > 0 Then
            f = Sound(ch * 4 + 2) + (Sound(ch * 4 + 3) And 7) * 256
            If f > 1 Then
                If ChannelWrite(ch) Then
                    ChannelWrite(ch) = False
                    LastFrame(ch) = Frames + l
                    playTone ch, getTone(f), v
                End If
            Else
                stopTone ch
            End If
        Else
            stopTone ch
        End If
    Else
        ChannelWrite(ch) = True
        stopTone ch
    End If
    If Frames >= LastFrame(ch) Then
        stopTone ch
    End If
End Sub

Public Sub PlayNoise(ch)
    Dim f, l, v
    If SoundCtrl And pow2(ch) Then
        v = 5
        l = vlengths(Sound(ch * 4 + 3) \ 8)
        If v > 0 Then
            f = (Sound(ch * 4 + 2) And 15) * 128
            If f > 1 Then
                If ChannelWrite(ch) Then
                    ChannelWrite(ch) = False
                    LastFrame(ch) = Frames + l
                    playTone ch, getTone(f), v
                End If
            Else
                stopTone ch
            End If
        Else
            stopTone ch
        End If
    Else
        ChannelWrite(ch) = True
        stopTone ch
    End If
    If Frames >= LastFrame(ch) Then
        stopTone ch
    End If
End Sub


Public Sub updateSounds()
    If doSound Then
        PlayRect 0
        PlayRect 1
        PlayTriangle 2
        PlayNoise 3
    Else
        stopTone 0
        stopTone 1
        stopTone 2
        stopTone 3
    End If
End Sub

#If 0 Then
Todo List:
*Mostly done*1) get it to go through sound() and play or stop sounds
*Done*2) get the tones sounding right
*In Progress*3) more scary stuff
#End If
